@@ -586,10 +586,9 @@ class JSONRPCConnection:
         instead.
     '''
 
-    _id_counter = itertools.count()
-
     def __init__(self, protocol):
         self._protocol = protocol
+        self._id_counter = itertools.count()
         # Sent Requests and Batches that have not received a response.
         # The key is its request ID; for a batch it is sorted tuple
         # of request IDs
@@ -611,10 +610,11 @@ class JSONRPCConnection:
                 message = f'response to unsent request (ID: {request_id})'
             raise ProtocolError.invalid_request(message) from None
         _request, future = self._requests.pop(request_id)
-        if isinstance(result, Exception):
-            future.set_exception(result)
-        else:
-            future.set_result(result)
+        if not future.done():
+            if isinstance(result, Exception):
+                future.set_exception(result)
+            else:
+                future.set_result(result)
         return []
 
     def _receive_request_batch(self, payloads):
@@ -665,7 +665,8 @@ class JSONRPCConnection:
         if ordered_ids not in self._requests:
             raise ProtocolError.invalid_request('response to unsent batch')
         _request_batch, future = self._requests.pop(ordered_ids)
-        future.set_result(ordered_results)
+        if not future.done():
+            future.set_result(ordered_results)
         return []
 
     def _send_result(self, request_id, result):
@@ -741,10 +742,9 @@ class JSONRPCConnection:
 
     def cancel_pending_requests(self):
         '''Cancel all pending requests.'''
-        exception = CancelledError()
         for _request, future in self._requests.values():
             if not future.done():
-                future.set_exception(exception)
+                future.cancel()
         self._requests.clear()
 
     def pending_requests(self):
